from django import template
from django.template import Context
from django.template import Library

from ..templatetags.react import ComponentNode
from ..templatetags.react import ComponentProps
from ..templatetags.react import parse_component_tag
from .utils import get_module_import_source


class LabeledInputNode(ComponentNode):
    def resolve_props(self, context: Context) -> ComponentProps:
        values = super().resolve_props(context)
        # LabeledInput doesn't support validationState, it only handles ``isInvalid`` so convert it here
        validation_state = values.pop("validationState", "")
        if validation_state == "invalid":
            values.update({"isInvalid": True})
        return values


def labeled_input(parser: template.base.Parser, token: template.base.Token):
    """
    Render ``LabeledInput``.

    For compatability with ``form_input`` it will convert ``validationState="invalid"`` into ``isInvalid=True``.
    """
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "LabeledInput", False, parser.origin),
        node_class=LabeledInputNode,
    )


def register_labeled_input(register: Library):
    register.tag("LabeledInput")(labeled_input)



from django import template
from django.template import Library

from ..templatetags.react import parse_component_tag
from .utils import get_module_import_source


def button(parser: template.base.Parser, token: template.base.Token):
    """
    Render a ``Button`` component from the Alliance UI React library with the specified props.

    If used for a link, the :function:`~alliance_platform.templatetags.alliance_ui.url_with_perms_filter`
    or :function:`~alliance_platform.frontend.templatetags.alliance_ui.url_filter` can be used to handle
    url args, or hide if the link is not available.
    """
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "Button", False, parser.origin),
    )


def button_group(parser: template.base.Parser, token: template.base.Token):
    """Render a ``ButtonGroup`` component from the Alliance UI React library with the specified props."""
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "ButtonGroup", False, parser.origin),
    )


def register_button(register: Library):
    register.tag("Button")(button)
    register.tag("ButtonGroup")(button_group)



from __future__ import annotations

from typing import Any

from allianceutils.auth.permission import reverse_if_probably_allowed
from django import template
from django.db.models import Model
from django.http import HttpRequest
from django.template import Context
from django.urls import reverse

from ..alliance_ui.button import register_button
from ..alliance_ui.date_picker import register_date_picker
from ..alliance_ui.icon import register_icon
from ..alliance_ui.inline_alert import register_inline_alert
from ..alliance_ui.labeled_input import register_labeled_input
from ..alliance_ui.menubar import register_menubar
from ..alliance_ui.misc import register_misc
from ..alliance_ui.pagination import register_pagination
from ..alliance_ui.table import register_table
from ..alliance_ui.time_input import register_time_input
from .react import DeferredProp
from .react import OmitComponentFromRendering

register = template.Library()

register_menubar(register)
register_table(register)
register_pagination(register)
register_button(register)
register_misc(register)
register_inline_alert(register)
register_icon(register)
register_date_picker(register)
register_time_input(register)
register_labeled_input(register)


class NamedUrlDeferredProp(DeferredProp):
    """Used by ``url_with_perm`` and ``url`` filters to defer the resolution of a URL until rendering time.

    This allows perm checks to occur and to omit the component from rendering if failed"""

    url_name: str
    check_perm: bool
    args: list[Any]
    kwargs: dict[str, Any]
    object: Model | None

    def __init__(self, url_name, check_perm=False):
        self.url_name = url_name
        self.check_perm = check_perm
        self.args = []
        self.kwargs = {}
        self.object = None
        super().__init__()

    def set_object(self, arg: Model):
        self.object = arg

    def add_arg(self, arg: Any):
        self.args.append(arg)

    def add_kwargs(self, kwargs: dict[str, Any]):
        self.kwargs.update(kwargs)

    def resolve(self, context: Context):
        if self.check_perm:
            url = reverse_if_probably_allowed(
                context["request"], self.url_name, object=self.object, args=self.args, kwargs=self.kwargs
            )
            if url is None:
                raise OmitComponentFromRendering()
            return url
        return reverse(self.url_name, args=self.args, kwargs=self.kwargs)


@register.filter("url_with_perm")
def url_with_perm_filter(value, arg1=None):
    """Resolve a URL and check if the current user has permission to access it.

    If permission check fails, the component that uses the value will be omitted from rendering.
    """
    pfv = NamedUrlDeferredProp(value, True)
    if arg1 is not None:
        pfv.add_arg(arg1)
    return pfv


@register.filter("url")
def url_filter(value, arg1=None):
    """Behaves same as ``url_with_perm`` but does not check any permissions."""
    pfv = NamedUrlDeferredProp(value)
    if arg1 is not None:
        pfv.add_arg(arg1)
    return pfv


@register.filter("with_arg")
def with_arg(value: NamedUrlDeferredProp, arg):
    """Add an argument to a ``url_with_perm`` or ``url`` filter."""
    value.add_arg(arg)
    return value


@register.filter("with_kwargs")
def with_kwargs(value: NamedUrlDeferredProp, kwargs):
    """Add kwargs to a ``url_with_perm`` or ``url`` filter."""
    value.add_kwargs(kwargs)
    return value


@register.filter("with_perm_obj")
def with_perm_object(value: NamedUrlDeferredProp, arg):
    """Add an object to a ``url_with_perm`` filter for the purposes of object level permission checks."""
    if not isinstance(value, NamedUrlDeferredProp):
        raise ValueError(
            '`perm_obj` must appear after `url_with_perm`, eg. `"url_name"|url_with_perm:obj.pk|with_perm_obj:obj'
        )
    value.set_object(arg)
    return value


@register.filter("unwrap_list")
def unwrap_list(value):
    if isinstance(value, (list, tuple)):
        if not value:
            return None
        if len(value) != 1:
            raise ValueError(f"Expected list of length 1, received length {len(value)}: {value}")
        return value[0]
    return value


@register.filter
def table_sort_order(request: HttpRequest, ordering_param: str = "ordering"):
    """Extract the current sort ordering from the request GET parameters and return it as a list of dicts

    Each ordering entry is a dict with keys "column" and "direction".

    For example, the URL /foo?ordering=-bar,email would return::

        [
          {'column': 'email', 'direction': 'ascending'},
          {'column': 'name', 'direction': 'descending'}
        ]
    """
    query = request.GET.copy()
    current_sorting_str = query.get(ordering_param)
    if current_sorting_str:
        return [
            {
                "column": x[1:] if x.startswith("-") else x,
                "direction": "descending" if x.startswith("-") else "ascending",
            }
            for x in current_sorting_str.split(",")
        ]
    return []



from __future__ import annotations

from django import template
from django.template import Library

from ..bundler import get_bundler
from ..bundler.base import ResolveContext
from ..templatetags.react import ImportComponentSource
from ..templatetags.react import parse_component_tag


def fragment_component(parser: template.base.Parser, token: template.base.Token):
    """Render a React Fragment component."""
    bundler = get_bundler()
    resolver_context = ResolveContext(bundler.root_dir, parser.origin.name if parser.origin else None)
    source_path = get_bundler().resolve_path(
        "/node_modules/react", resolver_context, resolve_extensions=[".ts", ".tsx", ".js"]
    )
    asset_source = ImportComponentSource(source_path, "React", True, "Fragment")
    return parse_component_tag(parser, token, asset_source=asset_source)


def register_misc(register: Library):
    register.tag("Fragment")(fragment_component)



from __future__ import annotations

from django import template
from django.template import Library

from ..templatetags.react import parse_component_tag
from .utils import get_module_import_source


def menubar(parser: template.base.Parser, token: template.base.Token):
    """Render a ``Menubar`` component from the Alliance UI library with the specified props"""
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "Menubar", False, parser.origin),
    )


def submenu(parser: template.base.Parser, token: template.base.Token):
    """Render a ``Menubar.SubMenu`` component from the Alliance UI library with the specified props"""
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source(
            "@alliancesoftware/ui", "Menubar", False, parser.origin, property_name="SubMenu"
        ),
    )


def menubar_item(parser: template.base.Parser, token: template.base.Token):
    """
    Render a ``Menubar.Item`` component from the Alliance UI library with the specified props

    If used for a link, the :function:`~alliance_platform.frontend.templatetags.alliance_ui.url_with_perms_filter``
    or :function:`~alliance_platform.templatetags.alliance_ui.url_filter`` can be used to handle
    url args, or hide if the link is not available.
    """
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "Item", False, parser.origin),
    )


def menubar_section(parser: template.base.Parser, token: template.base.Token):
    """Render a ``Menubar.Section`` component from the Alliance UI library with the specified props"""
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "Section", False, parser.origin),
    )


def register_menubar(register: Library):
    register.tag("Menubar")(menubar)
    register.tag("Menubar.SubMenu")(submenu)
    register.tag("Menubar.Item")(menubar_item)
    register.tag("Menubar.Section")(menubar_section)



from dataclasses import dataclass
from dataclasses import field
from html.parser import HTMLParser
import re
from typing import TYPE_CHECKING
from typing import Any
from typing import Union
import warnings

from alliance_platform.frontend.util import transform_attribute_names
from django.template import Context
from django.template import Node
from django.template import NodeList
from django.template import Origin

if TYPE_CHECKING:
    from alliance_platform.frontend.templatetags.react import ComponentNode

# These are used to replace template Nodes in a string with placeholders. The resulting string can them be parsed as HTML,
# and then the placeholders replaced with the original template nodes.
_html_replacement_placeholder_prefix = "_______placeholder_____$"
_html_replacement_placeholder_suffix = "$_"
html_replacement_placeholder_template = (
    _html_replacement_placeholder_prefix + "{0}" + _html_replacement_placeholder_suffix
)


@dataclass
class Element:
    name: str
    attrs: dict


@dataclass
class HTMLElement:
    tag: str
    attributes: dict[str, str] = field(default_factory=dict)
    children: list[Union["HTMLElement", str]] = field(default_factory=list)


# Extracted from https://developer.mozilla.org/en-US/docs/Glossary/Void_element
void_elements = [
    "area",
    "base",
    "br",
    "col",
    "embed",
    "hr",
    "img",
    "input",
    "link",
    "meta",
    "param",
    "source",
    "track",
    "wbr",
]

# Extracted from https://html.spec.whatwg.org/multipage/indices.html#attributes-3
boolean_html_attributes = [
    "allowfullscreen",
    "async",
    "autofocus",
    "autoplay",
    "checked",
    "controls",
    "default",
    "defer",
    "disabled",
    "formnovalidate",
    "inert",
    "ismap",
    "itemscope",
    "multiple",
    "muted",
    "nomodule",
    "novalidate",
    "open",
    "playsinline",
    "readonly",
    "required",
    "reversed",
    "selected",
    "shadowrootdelegatesfocus",
    "shadowrootclonable",
    "shadowrootserializable",
]


class HtmlTreeParser(HTMLParser):
    def __init__(self):
        super().__init__(convert_charrefs=True)
        # The root of the parsed tree. The tag is nonsense, but not used - it's just to make implementation easier. We
        # just use it for ``children``
        self.root = HTMLElement("Root")
        # Stack to maintain hierarchy of elements
        self.stack = [self.root]

    def handle_starttag(self, tag, attrs):
        element = HTMLElement(tag, {attr[0]: attr[1] for attr in attrs})
        if self.stack:
            self.stack[-1].children.append(element)
        else:
            self.root = element
        self.stack.append(element)
        # for void elements close them immediately
        if tag in void_elements:
            self.handle_endtag(tag)

    def handle_endtag(self, tag):
        if self.stack and self.stack[-1].tag == tag:
            self.stack.pop()

    def handle_data(self, data):
        if data and self.stack:
            self.stack[-1].children.append(data)


def is_valid_html_attribute_name(attr_name):
    # This isn't strictly accurate, but it's good enough for our purposes.
    # https://html.spec.whatwg.org/multipage/syntax.html#attributes-2
    pattern = r'^[^\s\/<>=\'"=]+$'
    return bool(re.match(pattern, attr_name))


class HtmlAttributeTemplateNodeList:
    """A list of HTML attributes that may contain template nodes.

    Take an HTML string like::

    .. code-block:: html+django

            <input {% if value %}{{ value }}{% endif %} {% include "attrs.html" }>

    When this is parsed by ``convert_html_string`` the attributes cannot be resolved until the context is available.
    This gives us some template nodes, in the above example an ``IfNode`` and ``IncludeNode``. At some point we
    want to resolve this attributes using the context. This class is used to store this and resolve it later.
    """

    def __init__(self, attrs: list[Node | str], original_html: str, origin: Origin):
        self.original_html = original_html
        self.attrs = attrs
        self.origin = origin

    def resolve(self, context: Context) -> dict[str, str]:
        attr_str = " ".join([prop.render(context) if isinstance(prop, Node) else prop for prop in self.attrs])
        parser = HtmlTreeParser()
        parser.feed(f"<div {attr_str}></div>")
        tags = parser.root.children
        if tags:
            return transform_html_attributes(tags[0].attributes, self.original_html, self.origin)
        return {}


def convert_html_string(
    html: str, origin: Origin, *, replacements: dict[str, Node] | None = None
) -> list[Union["ComponentNode", str]]:
    """
    Given a string that may contain HTML, convert it to a tree of ``ComponentNode``s

    Args:
        html: The html string to convert
        origin: The template origin
        replacements: Any replacements to make in the HTML after conversion. The way this works is that a template NodeList
            is processed into a string, with any template nodes replaced with placeholders. The string is converted to
            a component tree, then any placeholders are replaced with the original template nodes.
    Returns:

    """
    # NOTE: I tried lxml initially (with & without BeautifulSoup), and it was slower for our specific use case.
    # In general, it's considered very fast, but we don't need most of its features and this simple parser was
    # faster.
    parser = HtmlTreeParser()
    # Parse the HTML
    parser.feed(html)

    tags = parser.root.children

    def handle_placeholders(content: str):
        if not replacements:
            return [content]
        parts = content.split(_html_replacement_placeholder_prefix)
        first = parts.pop(0)
        if not parts:
            return [first]
        children: list[str | Node] = []
        if first:
            children.append(first)
        for part in parts:
            placeholder_id, extra = part.split(_html_replacement_placeholder_suffix)
            children.append(replacements[placeholder_id])
            if extra:
                children.append(extra)
        return children

    def convert_attributes(attrs: dict[str, str | Any]):
        html_attribute_template_nodes = []
        transformed = {}
        for key, value in attrs.items():
            if not value:
                html_attribute_template_nodes += handle_placeholders(key)
            else:
                parts = handle_placeholders(value)
                transformed[key] = parts[0] if len(parts) == 1 else NodeList(parts)
        return transform_html_attributes(transformed, html, origin), html_attribute_template_nodes

    def convert_tree(tree):
        from alliance_platform.frontend.templatetags.react import CommonComponentSource
        from alliance_platform.frontend.templatetags.react import ComponentNode

        children = []
        for el in tree:
            if isinstance(el, str):
                parts = handle_placeholders(el)
                children += parts
            else:
                attrs, html_attribute_template_nodes = convert_attributes(el.attributes)
                children.append(
                    ComponentNode(
                        origin,
                        CommonComponentSource(el.tag),
                        {**attrs, "children": convert_tree(el.children)},
                        html_attribute_template_nodes=HtmlAttributeTemplateNodeList(
                            html_attribute_template_nodes, html, origin
                        ),
                    )
                )
        return children

    return convert_tree(tags)


def transform_html_attributes(attrs: dict[str, str | None], original_html: str, origin: Origin):
    """Transform the attributes of an HTML tag into the final form we want

    This does the following:
    - If any ``value`` is ``None``, set it to ``True`` if it's a boolean html attribute otherwise empty string
    - If any key is invalid, remove it and warn
    - Transform the attribute names to match what React expects
    """

    bad_keys: list[str] = []
    final_attrs = {}
    for key, value in attrs.items():
        if not is_valid_html_attribute_name(key):
            bad_keys.append(key)
        elif value is None:
            # for boolean attributes set it to ``True``, otherwise empty string
            final_attrs[key] = True if key in boolean_html_attributes else ""
        else:
            final_attrs[key] = value
    if bad_keys:
        warnings.warn(
            f"{origin} contained invalid HTML in '{original_html}'. The following parts were removed from tag: {', '.join(bad_keys)}"
        )
    return transform_attribute_names(final_attrs)



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
import warnings

from allianceutils.template import parse_tag_arguments
from allianceutils.template import resolve
from django import template
from django.forms import BaseForm
from django.forms import BoundField
from django.template import Context
from django.template import Node
from django.template import NodeList
from django.template import Origin
from django.template import TemplateSyntaxError
from django.template.base import UNKNOWN_SOURCE
from django.template.base import FilterExpression

from ..alliance_ui.labeled_input import LabeledInputNode
from ..alliance_ui.utils import get_module_import_source
from ..forms.renderers import FormInputContextRenderer
from .react import ComponentNode
from .react import NestedComponentProp
from .react import NestedComponentPropAccumulator
from .react import convert_html_string

register = template.Library()


class _NOT_PROVIDED:
    pass


class FormInputNode(template.Node):
    origin: Origin
    show_valid_state: FilterExpression | bool
    help_text: FilterExpression | str | type[_NOT_PROVIDED]
    label: FilterExpression | str | type[_NOT_PROVIDED]
    is_required: FilterExpression | bool
    extra_attrs: dict[str, FilterExpression]

    def __init__(
        self,
        field: FilterExpression,
        origin: Origin | None,
        help_text=_NOT_PROVIDED,
        label=_NOT_PROVIDED,
        is_required=None,
        show_valid_state=True,
        non_standard_widget=False,
        **extra_attrs,
    ):
        self.origin = origin or Origin(UNKNOWN_SOURCE)
        self.show_valid_state = show_valid_state
        self.field = field
        self.help_text = help_text
        self.label = label
        self.required = is_required
        self.extra_attrs = extra_attrs or {}
        self.non_standard_widget = non_standard_widget

    def render(self, context: template.Context):
        field: BoundField = self.field.resolve(context)
        form_context = FormContext.get_current(context)
        extra_attrs = {}
        if not isinstance(field.form.renderer, FormInputContextRenderer):
            warnings.warn("form_input tag should only be used with 'FormInputContextRenderer'")
        else:
            help_text = (
                resolve(self.help_text, context) if self.help_text is not _NOT_PROVIDED else field.help_text
            )
            if help_text:
                # Help text can be HTML and django docs make it clear this value is not HTML-escaped.
                help_text = convert_html_string(help_text, self.origin)[0]
            extra_attrs[field.form.renderer.form_input_context_key] = {
                "raw_value": field.value(),
                "extra_widget_props": {
                    "label": resolve(self.label, context) if self.label is not _NOT_PROVIDED else field.label,
                    "errorMessage": ", ".join(str(e) for e in field.errors),
                    "validationState": (
                        "invalid"
                        if field.errors
                        else (
                            "valid"
                            if field.form.is_bound and resolve(self.show_valid_state, context)
                            else None
                        )
                    ),
                    "description": help_text,
                    "isRequired": (
                        resolve(self.required, context) if self.required is not None else field.field.required
                    ),
                    **{key: resolve(value, context) for key, value in self.extra_attrs.items()},
                    **form_context.get_extra_props(field),
                },
            }
            form_context.mark_processed(field.name)
            if self.non_standard_widget:
                node = LabeledInputNode(
                    self.origin,
                    get_module_import_source("@alliancesoftware/ui", "LabeledInput", False, self.origin),
                    {
                        **extra_attrs[field.form.renderer.form_input_context_key]["extra_widget_props"],
                        "children": [],
                    },
                )
                node.props.update({"children": NodeList([NestedComponentFormWidgetNode(field, self.origin)])})
                return node.render(context)
            if NestedComponentPropAccumulator.context_key in context:
                # If this is used within another React component, we need to defer rendering of the widget so that
                # it has access to the NestedComponentPropAccumulator
                return NestedComponentFormWidgetNode(field, self.origin, extra_attrs).render(context)
        return field.as_widget(attrs=extra_attrs)  # type: ignore[arg-type] # dict type in form says on str|bool but other types seem to work fine


def _replace_nodes_with_str(items: list[str | ComponentNode], separator: str):
    replacements: dict[str, ComponentNode] = {}
    final_items: list[str] = []
    for item in items:
        if isinstance(item, ComponentNode):
            placeholder = f"__form_widget_placeholder({len(replacements)})__"
            replacements[placeholder] = item
            final_items.append(placeholder)
        else:
            final_items.append(item)
    return separator.join(final_items), replacements


class NestedComponentFormWidgetNode(Node):
    field: BoundField
    origin: Origin
    attrs: dict

    def __init__(self, field: BoundField, origin: Origin, attrs: dict | None = None):
        self.field = field
        self.origin = origin
        self.attrs = attrs or {}

    def render(self, context: Context):
        """
        This is a bit of a hack, so some explanation is needed.

        The problem we are solving is that widgets are rendered as standalone templates. This means that
        the contents is not treated as a component child, even though in this case we are treating it as such.

        temp_accumulator is a temporary accumulator that is used to collect the children of the widget as
        it is rendered. Note that we have to pass this through under ``attrs`` to the ``as_widget`` method which
        makes it appear under ``widget.attrs`` in the final widget template context. This needs to be at the top
        level however, so we rely on ``FormInputContextRenderer`` to pop this value and insert it into the context.
        This is the only way to manipulate the final context a widget template gets rendered with, otherwise you
        can only change ``widget.attrs``. ``FormInputContextRenderer`` is checked in ``FormInputNode`` so it's
        guaranteed to be used if this code path gets executed.

        This is necessary because the widget may contain nested components that need to be
        rendered within the context of the parent component (i.e. we don't want the <script> tags etc, we
        want it to get added to the ``accumulator`` above.

        However, there could be other HTML mixed in with those components so we have to do it in two passes.
        The first pass renders the widgets, collects any nested components and replaces them with placeholders.
        The second passes the first output through ``convert_html_string`` to generate a list of components (where
        HTML tags are encountered), or strings otherwise.

        Take for example this template. It has a mixture of component nodes that will be handled in first pass,
        HTML handled in second pass, and plain strings that require no transformation:

            {% component "@alliancesoftware/ui" "TextInput" props=widget.attrs %}{% endcomponent %}
            Plain text <strong>Nested HTML</strong> More text <span>Extra</span>
            {% component "i" %}Another Component{% endcomponent %}
            End

        ``widget_html`` will be:

            __NestedComponentPropAccumulator__prop__0
            Plain text <strong>Nested HTML</strong> More text <span>Extra</span>
            __NestedComponentPropAccumulator__prop__1
            End

        ``convert_html_string`` will return:

            [
              '__NestedComponentPropAccumulator__prop__0\nPlain text ',
              ComponentNode(CommonComponentSource(name='strong'), {'children': ['Nested HTML']}),
              ' More text ',
              ComponentNode(CommonComponentSource(name='span'), {'children': ['Extra']}),
              '\n__NestedComponentPropAccumulator__prop__1\nEnd'
            ]

        Next we replace any `ComponentNode` with a placeholder (__form_widget_placeholder(0)__) using ``_replace_nodes_with_str``
        which generates a string like:

            __NestedComponentPropAccumulator__prop__0
            Plain text __form_widget_placeholder(0)__ More text __form_widget_placeholder(1)__
            __NestedComponentPropAccumulator__prop__1
            End

        This gets passed through ``temp_accumulator.apply`` which will replace any of the placeholders from first
        pass ("__NestedComponentPropAccumulator__prop__0" in the example). This results in:

            [
                NestedComponentProp(ImportComponentSource(TextInput), ComponentProps({..})),
                '\nPlain text __form_widget_placeholder(0)__ More text __form_widget_placeholder(1)__\n',
                 NestedComponentProp(CommonComponentSource(name='i'), ComponentProps({'children': 'Another Component'})),
                 '\nEnd'
            ]

        Finally, we iterate over this list. If a `NestedComponentProp` is encountered, it's added as a child to the
        original parent component with ``accumulator.add`` (this returns a string placeholder). If a string is
        encountered we have to replace any of the __form_widget_placeholder(i)__ placeholders. We do this by getting
        the original component for that placeholder, add it with ``accumulator.add`` and use the placeholder returned
        in place of the original. This is repeated until all placeholders are replaced. This gives us a final
        string like:

            __NestedComponentPropAccumulator__prop__0
            Plain text __NestedComponentPropAccumulator__prop__1 More text __NestedComponentPropAccumulator__prop__2
            __NestedComponentPropAccumulator__prop__3
            End

        This is what we return, and the parent component will replace any of the remaining placeholders as part of
        the normal ``ComponentNode`` rendering.
        """

        # This is the parent component accumulator we are rendered within. This ``render`` method is being called
        # from within ``ComponentNode`` within an active NestedComponentPropAccumulator context.
        accumulator = context.get(NestedComponentPropAccumulator.context_key)
        if not accumulator:
            raise ValueError("Unexpected: NestedComponentPropAccumulator not found in context")

        with NestedComponentPropAccumulator(context, accumulator.origin_node) as temp_accumulator:
            widget_html = self.field.as_widget(
                attrs={**self.attrs, NestedComponentPropAccumulator.context_key: temp_accumulator}
            )
            final = []
            # if this string appears in the string then it's going to be lost in the final output, but
            # considering they aren't visible characters it likely doesn't matter.
            # Zero-width Space + Information Separator Four
            separator = "\u200b\u001c"
            combined_str, replacements = _replace_nodes_with_str(
                convert_html_string(widget_html, self.origin), separator
            )
            for item in temp_accumulator.apply(combined_str):
                if isinstance(item, str):
                    for part in item.split(separator):
                        if part in replacements:
                            prop = NestedComponentProp(replacements[part], accumulator.origin_node, context)
                            final.append(accumulator.add(prop))
                        else:
                            final.append(part)
                else:
                    final.append(accumulator.add(item))
            return "".join(final)


@register.tag("form_input")
def form_input(parser: template.base.Parser, token: template.base.Token):
    """Renders a form input with additional props supported by alliance-ui

    This sets `label`, `errorMessage`, `validationState`, `description` and `isRequired`. In addition, it may
    set `autoFocus` based on the `auto_focus` setting on the parent `form` tag.

    The following options can be passed to the tag to override defaults:

    - `label` - set the label for the input. If not specified will use ``field.label``.
    - `help_text` - help text to show below the input. If not specified will use ``field.help_text``.
    - `show_valid_state` - if true, will show the 'valid' (i.e. success) state of the input. If not specified will use
      ``False``. For most components in alliance-ui this results in it showing a tick icon and/or rendering green. If
      this is ``False`` only error states will be shown.
    - `is_required` - if true, will show the input as required. If not specified will use the model field ``required``
      setting.

    In addition, you can pass through any extra attributes that should be set on the input. For example, to set an
    addon for an alliance-ui ``TextInput`` you could do the following::

        {% form_input field addonBefore="$" %}

    Note that the attributes supported here depend entirely on the widget. If the widget is a React component, you
    can also pass react components to the tag::

        {% component "core-ui/icons" "Search" as search_icon %}${% endcomponent %}
        {% form_input field addonBefore=search_icon %}

    The additional props are added to the key ``extra_widget_props`` - so the relevant widget template needs to include
    this for the props to be passed through::

        {% component "@alliancesoftware/ui" "TextInput" props=widget.attrs|merge_props:extra_widget_props|html_attr_to_jsx type=widget.type name=widget.name default_value=widget.value %}
        {% endcomponent %}
    """
    tag_name = token.split_contents()[0]
    args, kwargs, target_var = parse_tag_arguments(parser, token, supports_as=True)
    if not len(args) == 1:
        raise TemplateSyntaxError(f"{tag_name} must be passed the form field to render input for")

    return FormInputNode(
        args[0],
        parser.origin,
        **kwargs,
    )


@dataclass
class FormContext:
    form: BaseForm
    #: If true, the first field with errors will be focused or the first field if form not submitted
    auto_focus: bool = False
    processed_fields: list[str] = field(default_factory=list)

    def mark_processed(self, field_name: str):
        self.processed_fields.append(field_name)

    @classmethod
    def get_current(cls, context: template.Context) -> FormContext:
        form_context = context.get("form_context", None)
        if not form_context:
            raise ValueError("form_input tag must be used within a form tag")
        return form_context

    _focused_field: BoundField | None = None

    def get_extra_props(self, field: BoundField):
        extra_props = {}
        if self.auto_focus:
            if self.form.is_bound and field.errors and not self._focused_field:
                self._focused_field = field
                extra_props["auto_focus"] = True
            elif not self.form.is_bound and not self.processed_fields:
                # focus first field
                extra_props["auto_focus"] = True
        return extra_props


class FormNode(template.Node):
    origin: Origin
    form: FilterExpression
    nodelist: NodeList
    auto_focus: FilterExpression | bool

    def __init__(self, form: FilterExpression, nodelist: NodeList, origin: Origin | None, auto_focus=False):
        self.origin = origin or Origin(UNKNOWN_SOURCE)
        self.form = form
        self.nodelist = nodelist
        self.auto_focus = auto_focus

    def render(self, context: template.Context):
        form: BaseForm = resolve(self.form, context)
        with context.push(form_context=FormContext(form, resolve(self.auto_focus, context))):
            return self.nodelist.render(context)


@register.tag("form")
def form(parser: template.base.Parser, token: template.base.Token):
    """Tag to setup a form context for form_input tags

    This tag doesn't render anything itself, it just sets up context for ``form_input`` tags. This is to support
    the ``auto_focus`` behaviour. This works by adding an ``auto_focus`` prop to the first field with errors, or the
    first rendered field if no errors are present.

    Usage::

        {% load form %}

        {% form form auto_focus=True %}
            <form method="post>
            {% for field in form.visible_fields %}
              {% form_input field %}
            {% endfor %}
            </form>
        {% endform %}


    .. note::

        Usage of this tag requires the following settings to be set::

            FORM_RENDERER = "alliance_platform.frontend.forms.renderers.FormInputContextRenderer"

    """
    tag_name = token.split_contents()[0]
    args, kwargs, target_var = parse_tag_arguments(parser, token, supports_as=True)
    if not len(args) == 1:
        raise TemplateSyntaxError(f"{tag_name} must be passed the django form instance")

    nodelist = parser.parse((f"end{tag_name}",))
    parser.delete_first_token()

    return FormNode(
        args[0],
        nodelist,
        parser.origin,
        **kwargs,
    )
